from arcgis._impl.common._mixins import PropertyMap
from collections import OrderedDict
from arcgis.geocoding import geocode
import json


class Event(OrderedDict):
    """
    Represents an event in a Hub. A Hub has many Events that can be associated with an Initiative.
    Events are meetings for people to support an Initiative. Events are scheduled by an organizer
    and have many attendees. An Event has a Group so that they can include content for preparation
    as well as gather and archive content during the event for later retrieval or analysis.
    """

    def __init__(self, gis, event_object):
        """
        Constructs an empty Event object
        """
        self._gis = gis
        self._hub = self._gis.hub
        self._eventdict = event_object["attributes"]
        try:
            self._eventdict["geometry"] = event_object["geometry"]
        except KeyError:
            self._eventdict["geometry"] = {"x": 0.00, "y": 0.00}
        pmap = PropertyMap(self._eventdict)
        self.definition = pmap

    def __repr__(self):
        return '<%s title:"%s" venue:%s>' % (
            type(self).__name__,
            self.title,
            self.venue,
        )

    @property
    def event_id(self) -> str:
        """
        Returns the unique identifier of the event
        """
        return self._eventdict["OBJECTID"]

    @property
    def title(self) -> str:
        """
        Returns the title of the event
        """
        return self._eventdict["title"]

    @property
    def venue(self) -> str:
        """
        Returns the location of the event
        """
        return self._eventdict["venue"]

    @property
    def address(self) -> str:
        """
        Returns the street address for the venue of the event
        """
        return self._eventdict["address1"]

    @property
    def initiative_id(self) -> str:
        """
        Returns the initiative id of the initiative the event belongs to
        """
        return self._eventdict["initiativeId"]

    @property
    def organizers(self) -> str:
        """
        Returns the name and email of the event organizers
        """
        return self._eventdict["organizers"]

    @property
    def description(self) -> str:
        """
        Returns description of the event
        """
        return self._eventdict["description"]

    @property
    def start_date(self) -> str:
        """
        Returns start date of the event in milliseconds since UNIX epoch
        """
        return self._eventdict["startDate"]

    @property
    def end_date(self) -> str:
        """
        Returns end date of the event in milliseconds since UNIX epoch
        """
        return self._eventdict["endDate"]

    @property
    def creator(self) -> str:
        """
        Returns creator of the event
        """
        return self._eventdict["Creator"]

    @property
    def capacity(self) -> int:
        """
        Returns attendance capacity for attendees of the event
        """
        return self._eventdict["capacity"]

    @property
    def attendance(self) -> int:
        """
        Returns attendance count for a past event
        """
        return self._eventdict["attendance"]

    @property
    def access(self) -> str:
        """
        Returns access permissions of the event
        """
        return self._eventdict["status"]

    @property
    def group_id(self) -> str:
        """
        Returns groupId for the event
        """
        return self._eventdict["groupId"]

    @property
    def is_cancelled(self) -> bool:
        """
        Check if event is Cancelled
        """
        return self._eventdict["isCancelled"]

    @property
    def geometry(self) -> dict:
        """
        Returns co-ordinates of the event location
        """
        return self._eventdict["geometry"]

    def delete(self) -> bool:
        """
        Deletes an event

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Delete an event successfully

            event1 = myhub.events.get(24)
            event1.delete()

            >> True
        """
        _group = self._gis.groups.get(self.group_id)
        _group.protected = False
        _group.delete()
        params = {
            "f": "json",
            "objectIds": self.event_id,
            "token": self._gis._con.token,
        }
        delete_event = self._gis._con.post(
            path="https://hub.arcgis.com/api/v3/events/"
            + self._hub.enterprise_org_id
            + "/Hub Events/FeatureServer/0/deleteFeatures",
            postdata=params,
        )
        return delete_event["deleteResults"][0]["success"]

    def update(self, event_properties: dict) -> bool:
        """
        Updates properties of an event

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Update an event successfully

            event1 = myhub.events.get(id)
            event_properties = {'status': 'planned', description: 'Test'}
            event1.update(event_properties)

            >> True
        """
        _feature = {}

        # Build event feature
        event_properties["OBJECTID"] = self.event_id
        _feature["attributes"] = self._eventdict
        for key, value in event_properties.items():
            _feature["attributes"][key] = value
        _feature["geometry"] = self.geometry
        event_data = [_feature]

        # Update event
        url = (
            "https://hub.arcgis.com/api/v3/events/"
            + self._hub.enterprise_org_id
            + "/Hub Events/FeatureServer/0/updateFeatures"
        )
        params = {"f": "json", "features": event_data, "token": self._gis._con.token}
        update_event = self._gis._con.post(path=url, postdata=params)
        return update_event["updateResults"][0]["success"]


class EventManager(object):
    """Helper class for managing events within a Hub. This class is not created by users directly.
    An instance of this class, called 'events', is available as a property of the Hub object. Users
    call methods on this 'events' object to manipulate (add, search, get_map etc) events
    of a particular Hub.
    """

    def __init__(self, hub, event=None):
        self._hub = hub
        self._gis = self._hub.gis
        if event:
            self._event = event

    def _all_events(self):
        """
        Fetches all events for particular hub.
        """
        events = []
        url = (
            "https://hub.arcgis.com/api/v3/events/"
            + self._hub.enterprise_org_id
            + "/Hub Events/FeatureServer/0/query"
        )
        params = {
            "f": "json",
            "outFields": "*",
            "where": "1=1",
            "token": self._gis._con.token,
        }
        all_events = self._gis._con.get(url, params)
        _events_data = all_events["features"]
        for event in _events_data:
            events.append(Event(self._gis, event))
        return events

    def add(self, event_properties: dict) -> Event:
        """
        Adds an event for an initiative.

        =================     ====================================================================
        **Parameter**          **Description**
        -----------------     --------------------------------------------------------------------
        event_properties      Required dictionary. See table below for the keys and values.
        =================     ====================================================================


        *Key:Value Dictionary Options for Argument event_properties*

        =================       =====================================================================
        **Key**                 **Value**
        -----------------       ---------------------------------------------------------------------
        title                   Required string. Name of event.
        -----------------       ---------------------------------------------------------------------
        description             Required string. Description of the event.
        -----------------       ---------------------------------------------------------------------
        initiaitve_id           Required string. Name label of the item.
        -----------------       ---------------------------------------------------------------------
        venue                   Required string. Venue name for the event.
        -----------------       ---------------------------------------------------------------------
        address1                Required string. Street address for the venue.
        -----------------       ---------------------------------------------------------------------
        status                  Required string. Access of event. Valid values are private, planned,
                                public, draft.
        -----------------       ---------------------------------------------------------------------
        startDate               Required start date of the event in milliseconds since UNIX epoch.
        -----------------       ---------------------------------------------------------------------
        endDate                 Required end date of the event in milliseconds since UNIX epoch.
        -----------------       ---------------------------------------------------------------------
        isAllDay                Required boolean. Indicates if the event is a day long event.
        -----------------       ---------------------------------------------------------------------
        capacity                Optional integer. The attendance capacity of the event venue.
        -----------------       ---------------------------------------------------------------------
        address2                Optional string.  Additional information about event venue street address.
        -----------------       ---------------------------------------------------------------------
        onlineLocation          Optional string. Web URL or other details for online event.
        -----------------       ---------------------------------------------------------------------
        organizers              Optional list of dictionary of keys `name` and `contact` for each organizer's
                                name and email. Default values are name, email, username of event creator.
        -----------------       ---------------------------------------------------------------------
        sponsors                Optional list of dictionary of keys `name` and `contact` for each sponsor's
                                name and contact.
        =================       =====================================================================

        :return:
            Event if successfully added.

        .. code-block:: python

            USAGE EXAMPLE: Add an event successfully

            event_properties = {
                'title':'Test Event',
                'description': 'Testing with python',
                'initiativeId': '43f..',
                'venue': 'Washington Monument',
                'address1': '2 15th St NW, Washington, District of Columbia, 20024',
                'status': 'planned',
                'startDate': 1562803200,
                'endDate': 1562889600,
                'isAllDay': 1
            }

            new_event = myhub.events.add(event_properties)
        """
        _feature = {}
        # Fetch initiaitve site id
        _initiative = self._hub.initiatives.get(event_properties["initiativeId"])
        event_properties["siteId"] = _initiative.site_id
        # Set organizers if not provided
        try:
            event_properties["organizers"]
        except:
            _organizers_list = [
                {
                    "name": self._gis.users.me.fullName,
                    "contact": self._gis.users.me.email,
                    "username": self._gis.users.me.username,
                }
            ]
            _organizers = json.dumps(_organizers_list)
            event_properties["organizers"] = _organizers
        # Set sponsors if not provided
        try:
            event_properties["sponsors"]
            event_properties["sponsors"] = json.dumps(event_properties["sponsors"])
        except:
            _sponsors = []
            event_properties["sponsors"] = json.dumps(_sponsors)
        # Set onlineLocation if not provided
        try:
            event_properties["onlineLocation"]
        except:
            _onlineLocation = ""
            event_properties["onlineLocation"] = _onlineLocation
        # Set geometry if not provided
        try:
            event_properties["geometry"]
            geometry = event_properties["geometry"]
            del event_properties["geometry"]
        except:
            geometry = geocode(event_properties["address1"])[0]["location"]

        event_properties["schemaVersion"] = 2
        event_properties["location"] = ""
        event_properties["url"] = event_properties["title"].replace(" ", "-").lower()

        # Generate event id for new event
        event_id = max([event.event_id for event in self._all_events()]) + 1

        # Create event group
        _event_group_dict = {
            "title": event_properties["title"],
            "access": "public",
            "tags": ["Hub Event Group", "Open Data", "hubEvent|" + str(event_id)],
        }
        _event_group = self._gis.groups.create_from_dict(_event_group_dict)
        _event_group.protected = True
        event_properties["groupId"] = _event_group.id

        # Build new event feature and create it
        _feature["attributes"] = event_properties
        _feature["geometry"] = geometry
        event_data = [_feature]
        url = (
            "https://hub.arcgis.com/api/v3/events/"
            + self._hub.enterprise_org_id
            + "/Hub Events/FeatureServer/0/addFeatures"
        )
        params = {"f": "json", "features": event_data, "token": self._gis._con.token}
        add_event = self._gis._con.post(path=url, postdata=params)
        try:
            add_event["addResults"]
            return self.get(add_event["addResults"][0]["objectId"])
        except:
            return add_event

    def search(self, initiative_id=None, title=None, venue=None, organizer_name=None):
        """
        Searches for events within a Hub.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        initiative_id       Optional string. Initiative itemid.
        ---------------     --------------------------------------------------------------------
        title               Optional string. Title of the event.
        ---------------     --------------------------------------------------------------------
        venue               Optional string. Venue where event is held.
        ---------------     --------------------------------------------------------------------
        organizer_name      Optional string. Name of the organizer of the event.
        ===============     ====================================================================

        :return:
           A list of matching indicators.

        """
        events = []
        events = self._all_events()
        if initiative_id != None:
            # events =
            events = [event for event in events if initiative_id == event.initiative_id]
        if title != None:
            events = [event for event in events if title in event.title]
        if venue != None:
            events = [event for event in events if venue in event.venue]
        if organizer_name != None:
            events = [event for event in events if organizer_name in event.organizers]
        return events

    def get(self, event_id: int) -> Event:
        """Get the event for the specified event_id.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        event_id                   Required integer. The event identifier.
        =======================    =============================================================

        :return:
            The :class:`~arcgis.apps.hub.Event` object.

        """
        url = (
            "https://hub.arcgis.com/api/v3/events/"
            + self._hub.enterprise_org_id
            + "/Hub Events/FeatureServer/0/"
            + str(event_id)
        )
        params = {"f": "json", "token": self._gis._con.token}
        feature = self._gis._con.get(url, params)
        return Event(self._gis, feature["feature"])



"""
Defines the AssignmentType class.
"""
from .exceptions import ValidationError
from .feature_model import FeatureModel
from ._store import *
from ._store.assignment_types_v2 import *
from ._schemas import AssignmentTypeSchema


class AssignmentType(FeatureModel):
    """
    Defines the acceptable values for :class:`~arcgis.apps.workforce.AssignmentType` types.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                           this assignment belongs to.
    ------------------     --------------------------------------------------------------------
    coded_value            Optional :class:`dict`. The dictionary storing the code and
                           name of the type. Only works for v1 projects.
    ------------------     --------------------------------------------------------------------
    name                   Optional :class:`String`. The name of the assignment type.
    ==================     ====================================================================

    .. code-block:: python

        # Get an assignment type, update it, delete it

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        assignment_type = project.assignment_types.search()[0]
        assignment_type.update(name="Manhole Inspection")
        assignment_type.delete()


    """

    def __init__(self, project, feature=None, coded_value=None, name=None):
        if project._is_v2_project:
            super().__init__(
                project=project,
                feature_layer=project.assignment_types_table,
                feature=feature,
            )
            self._schema = AssignmentTypeSchema(project.assignment_types_table)
            self._coded_value = None
            if not feature:
                self.name = name
        else:
            super().__init__()
            if coded_value:
                self._coded_value = coded_value
            else:
                self._coded_value = {"code": None, "name": name}
        self.project = project

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<AssignmentType {}>".format(self.name)

    def update(self, name=None):
        """
        Updates the assignment type on the server

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Optional :class:`String`.
                               The name of the assignment type
        ==================     ====================================================================
        """
        if self.project._is_v2_project:
            update_assignment_type_v2(self.project, self, name)
        else:
            update_assignment_type(self.project, self, name)

    def delete(self):
        """Deletes the assignment type from the server"""
        if self.project._is_v2_project:
            delete_assignment_types_v2(self.project, [self])
        else:
            delete_assignment_types(self.project, [self])

    @property
    def id(self):
        """Gets the id of the assignment type"""
        return self.code

    @property
    def code(self):
        """Gets the internal code that uniquely identifies the assignment type"""
        if not self.project._is_v2_project:
            return self._coded_value["code"]
        elif self._feature.attributes.get(self._schema.global_id):
            return self._feature.attributes.get(self._schema.global_id).upper()
        else:
            return None

    @property
    def name(self):
        """Gets/Sets The name of the assignment type"""
        if not self.project._is_v2_project:
            return self._coded_value["name"]
        else:
            return self._feature.attributes.get(self._schema.description)

    @name.setter
    def name(self, value):
        if not self.project._is_v2_project:
            self._coded_value["name"] = value
        else:
            self._feature.attributes[self._schema.description] = value

    @property
    def coded_value(self):
        """Gets the coded value"""
        return self._coded_value

    def _validate(self, **kwargs):
        errors = super()._validate(**kwargs)
        errors += self._validate_name()
        errors += self._validate_name_uniqueness(**kwargs)
        return errors

    def _validate_for_update(self, **kwargs):
        if not self.project._is_v2_project:
            return super()._validate_for_update(**kwargs) + self._validate_code()
        else:
            return super()._validate_for_update(**kwargs)

    def _validate_for_remove(self, **kwargs):
        if not self.project._is_v2_project:
            assignments = kwargs["assignments"]
            errors = super()._validate_for_remove(**kwargs) + self._validate_code()
            if assignments is None:
                schema = self.project._assignment_schema
                where = "{}={}".format(schema.assignment_type, self.code)
                assignments = self.project.assignments.search(where=where)
            else:
                assignments = [
                    a for a in assignments if a.assignment_type.code == self.code
                ]
        else:
            assignments = self.project.assignments.search(where="1=1")
            errors = super()._validate_for_remove(**kwargs)
            assignments = [
                a for a in assignments if a.assignment_type.code == self.code
            ]

        if assignments:
            errors.append(
                ValidationError("Cannot remove an in-use AssignmentType", self)
            )
        return errors

    def _validate_name(self):
        errors = []
        if self.name is None or self.name.isspace():
            errors.append(ValidationError("AssignmentType must have a name", self))
        elif ">" in self.name or "<" in self.name or "%" in self.name:
            errors.append(
                ValidationError("AssignmentType name contains invalid characters", self)
            )
        return errors

    def _validate_name_uniqueness(self, assignment_types=None):
        errors = []
        if assignment_types is None:
            assignment_types = self.project.assignment_types.search()
        for assignment_type in assignment_types:
            # note that code is AT guid for v2 projects
            if (
                self.name
                and assignment_type.name.lower() == self.name.lower()
                and assignment_type.code != self.code
            ):
                errors.append(
                    ValidationError("AssignmentType name must be unique", self)
                )
        return errors

    def _validate_code(self):
        errors = []
        if not isinstance(self.code, int):
            errors.append(ValidationError("Code must be a unique integer", self))
        return errors



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="EncoderDecoder",
    pretrained="torchvision://resnet101",
    backbone=dict(
        type="ResNet",
        depth=101,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        dilations=(1, 1, 2, 4),
        strides=(1, 2, 1, 1),
        norm_cfg=norm_cfg,
        norm_eval=False,
        style="pytorch",
        contract_dilation=True,
    ),
    decode_head=dict(
        type="ASPPHead",
        in_channels=2048,
        in_index=3,
        channels=512,
        dilations=(1, 12, 24, 36),
        dropout_ratio=0.1,
        num_classes=19,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
    ),
    auxiliary_head=dict(
        type="FCNHead",
        in_channels=1024,
        in_index=2,
        channels=256,
        num_convs=1,
        concat_input=False,
        dropout_ratio=0.1,
        num_classes=19,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=0.4),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/deeplabv3/deeplabv3_r101b-d8_512x1024_80k_cityscapes/deeplabv3_r101b-d8_512x1024_80k_cityscapes_20201226_171821-8fd49503.pth"



"""
The arcgis.network module contains classes and functions for network analysis. Network layers and analysis can be used
for operations such as finding the closest facility, the best route for a vehicle, the best routes for a fleet of
vehicles, locating facilities using location allocation, calculating an OD cost matrix, and generating service areas.
"""

from ._layer import (
    NetworkLayer,
    NetworkDataset,
    ClosestFacilityLayer,
    ServiceAreaLayer,
    RouteLayer,
    NAJob,
    ODCostMatrixLayer,
)

from arcgis.auth.tools import LazyLoader

analysis = LazyLoader("arcgis.network.analysis")



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# model settings
norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="EncoderDecoder",
    backbone=dict(
        type="FastSCNN",
        downsample_dw_channels=(32, 48),
        global_in_channels=64,
        global_block_channels=(64, 96, 128),
        global_block_strides=(2, 2, 1),
        global_out_channels=128,
        higher_in_channels=64,
        lower_in_channels=128,
        fusion_out_channels=128,
        out_indices=(0, 1, 2),
        norm_cfg=norm_cfg,
        align_corners=False,
    ),
    decode_head=dict(
        type="DepthwiseSeparableFCNHead",
        in_channels=128,
        channels=128,
        concat_input=False,
        num_classes=19,
        in_index=-1,
        norm_cfg=norm_cfg,
        align_corners=False,
        loss_decode=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=0.4),
    ),
    auxiliary_head=[
        dict(
            type="FCNHead",
            in_channels=128,
            channels=32,
            num_convs=1,
            num_classes=19,
            in_index=-2,
            norm_cfg=norm_cfg,
            concat_input=False,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=True, loss_weight=0.4
            ),
        ),
        dict(
            type="FCNHead",
            in_channels=64,
            channels=32,
            num_convs=1,
            num_classes=19,
            in_index=-3,
            norm_cfg=norm_cfg,
            concat_input=False,
            align_corners=False,
            loss_decode=dict(
                type="CrossEntropyLoss", use_sigmoid=True, loss_weight=0.4
            ),
        ),
    ],
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)



import os, sys
import tempfile
from functools import partial
from pathlib import Path
import json
import warnings
import traceback
from ..models._arcgis_model import ArcGISModel, model_characteristics_folder

HAS_FASTAI = True

try:
    import transformers
    import torch
    import torch.nn as nn
    import pandas as pd
    from fastai.text.data import NumericalizeProcessor, TokenizeProcessor
    from fastai.text.transform import Tokenizer
    from fastprogress.fastprogress import progress_bar
    from fastai.basic_train import Learner, DatasetType
    from fastai.train import to_fp16
    from fastai.metrics import accuracy
    from transformers import AutoTokenizer, AutoConfig
    from .._utils.text_data import (
        TextDataObject,
        save_data_in_model_metrics_html,
        copy_metrics,
    )
    from .._utils._seq2seq_utils import (
        SequenceToSequenceLearner,
        CorpusBLEU,
        seq2seq_loss,
        seq2seq_acc,
    )
    from .._utils.text_transforms import TransformersBaseTokenizer, TransformersVocab
    from ._arcgis_transformer import ModelBackbone, infer_model_type
    from .._utils.common import _get_emd_path
    from ._transformer_seq2seq import (
        TransformerForSequenceToSequence,
        backbone_models_reverse_map,
        transformer_architectures,
        transformer_seq_length,
    )
    from transformers import logging
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False
    from ._transformer_seq2seq import transformer_architectures, transformer_seq_length
else:
    warnings.filterwarnings("ignore", category=UserWarning, module="fastai")


class SequenceToSequence(ArcGISModel):
    """
    Creates a :class:`~arcgis.learn.text.SequenceToSequence` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required text data object, returned from
                            :class:`~arcgis.learn.prepare_textdata` function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Specifying the HuggingFace
                            transformer model name to be used to train the
                            model. Default set to 't5-base'.

                            To learn more about the available models or
                            choose models that are suitable for your dataset,
                            kindly visit:- https://huggingface.co/transformers/pretrained_models.html
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    verbose                 Optional string. Default set to `error`. The
                            log level you want to set. It means the amount
                            of information you want to display while training
                            or calling the various methods of this class.
                            Allowed values are - `debug`, `info`, `warning`,
                            `error` and `critical`.
    ---------------------   -------------------------------------------
    seq_len                 Optional Integer. Default set to 512. Maximum
                            sequence length (at sub-word level after tokenization)
                            of the training data to be considered for training
                            the model.
    ---------------------   -------------------------------------------
    mixed_precision         Optional Bool. Default set to False. If set
                            True, then mixed precision training is used
                            to train the model
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.SequenceToSequence` model object for sequence_translation task.
    """

    # supported transformer backbones
    supported_backbones = transformer_architectures

    def __init__(self, data, backbone="t5-base", **kwargs):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)

        self.logger = logging.get_logger()
        if kwargs.get("verbose", None):
            self.logger.setLevel(kwargs.get("verbose").upper())
        else:
            self.logger.setLevel(logging.ERROR)

        model_backbone = ModelBackbone(backbone)
        super().__init__(data, model_backbone)
        self._mixed_precision = kwargs.get("mixed_precision", False)
        self._seq_len = kwargs.get("seq_len", transformer_seq_length)
        self.shap_values = None
        self._create_text_learner_object(
            data,
            backbone,
            kwargs.get("pretrained_path", None),
            mixed_precision=self._mixed_precision,
            seq_len=self._seq_len,
        )

        self.learn.model = self.learn.model.to(self._device)
        layer_groups = self.learn.model.get_layer_groups()
        self.learn.split(layer_groups)
        # self._freeze()

    def _create_text_learner_object(
        self,
        data,
        backbone,
        pretrained_path=None,
        mixed_precision=False,
        seq_len=transformer_seq_length,
    ):
        self._model_type = infer_model_type(backbone, transformer_architectures)
        self.logger.info(f"Inferred Backbone: {self._model_type}")
        pretrained_model_name = backbone
        transformer_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)
        pad_first = True if transformer_tokenizer.padding_side == "left" else False
        pad_idx = transformer_tokenizer.pad_token_id
        base_tokenizer = TransformersBaseTokenizer(
            pretrained_tokenizer=transformer_tokenizer, seq_len=seq_len
        )
        tokenizer = Tokenizer(tok_func=base_tokenizer, pre_rules=[], post_rules=[])
        if sys.platform == "win32":
            tokenizer.n_cpus = 1
        vocab = TransformersVocab(tokenizer=transformer_tokenizer)
        numericalize_processor = NumericalizeProcessor(vocab=vocab)
        tokenize_processor = TokenizeProcessor(
            tokenizer=tokenizer, include_bos=False, include_eos=False
        )
        transformer_processor = [tokenize_processor, numericalize_processor]
        if data._is_empty or data._backbone != backbone:
            self.logger.info("Creating DataBunch")
            data._prepare_seq2seq_databunch(
                transformer_processor=transformer_processor,
                pad_first=pad_first,
                pad_idx=pad_idx,
                model_type=self._model_type,
                backbone=backbone,
            )
        databunch = data.get_databunch()

        config = AutoConfig.from_pretrained(pretrained_model_name)

        if pretrained_path is not None:
            pretrained_path = str(_get_emd_path(pretrained_path))

        model = TransformerForSequenceToSequence(
            architecture=self._model_type,
            pretrained_model_name=pretrained_model_name,
            config=config,
            pretrained_model_path=pretrained_path,
            seq_len=seq_len,
        )
        from IPython.utils import io

        with io.capture_output() as captured:
            model.init_model()  # output will be cleared
        n_y_vocab = model._config.vocab_size
        metrics = [seq2seq_acc, CorpusBLEU(n_y_vocab)]
        # metrics=[accuracy, CorpusBLEU(n_y_vocab)]
        loss_func = seq2seq_loss
        from fastai.layers import LabelSmoothingCrossEntropy, FlattenedLoss

        # loss_func = FlattenedLoss(LabelSmoothingCrossEntropy, axis=-1)

        self.learn = SequenceToSequenceLearner(
            databunch, model, metrics=metrics, loss_func=loss_func, path=data.path
        )

        if pretrained_path is not None:
            self.load(pretrained_path)

        if mixed_precision:
            if self._model_type in ["t5"]:
                error_message = (
                    f"Mixed precision training is not supported for transformer model - {self._model_type.upper()}."
                    "\nKindly turn off the `mixed_precision` flag to use this model in its default mode,"
                    f" or choose a different transformer architectures from - {transformer_architectures}"
                )
                raise Exception(error_message)
            self.logger.info("Converting model to 16 Bit Floating Point precision")
            self.learn = to_fp16(self.learn)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "seq2seq_acc", "corpus_bleu"]

    @classmethod
    def available_backbone_models(cls, architecture):
        """
        Get available models for the given transformer backbone

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        architecture            Required string. name of the transformer
                                backbone one wish to use. To learn more about
                                the available models or choose models that are
                                suitable for your dataset, kindly visit:-
                                https://huggingface.co/transformers/pretrained_models.html
        =====================   ===========================================

        :return: a tuple containing the available models for the given transformer backbone
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)
        return TransformerForSequenceToSequence._available_backbone_models(architecture)

    def freeze(self):
        """
        Freeze up to last layer group to train only the last layer group of the model.
        """
        self.learn.freeze()

    @classmethod
    def from_model(cls, emd_path, data=None, **kwargs):
        """
        Creates an SequenceToSequence model object from a Deep Learning
        Package(DLPK) or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Optional fastai Databunch. Returned data
                                object from :class:`~arcgis.learn.prepare_textdata` function or None for
                                inferencing.
                                Default value: None
        =====================   ===========================================

        :return: :class:`~arcgis.learn.text.SequenceToSequence` Object
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error

            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        pretrained_model = emd["PretrainedModel"]
        mixed_precision = emd["MixedPrecisionTraining"]
        text_cols = emd["TextColumns"]
        label_cols = emd["LabelColumns"]
        seq_len = emd.get("SequenceLength", transformer_seq_length)

        data_is_none = False
        if data is None:
            data_is_none = True
            data = TextDataObject(task="sequence_translation")
            data._backbone = pretrained_model
            data.create_empty_seq2seq_data(text_cols, label_cols)
            data.emd, data.emd_path = emd, emd_path.parent
        cls_object = cls(
            data,
            pretrained_model,
            pretrained_path=str(emd_path),
            mixed_precision=mixed_precision,
            seq_len=seq_len,
            **kwargs,
        )
        if data_is_none:
            cls_object._data._is_empty = True
        return cls_object

    def load(self, name_or_path):
        """
        Loads a saved SequenceToSequence model from disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        if "\\" in str(name_or_path) or "/" in str(name_or_path):
            name_or_path = str(_get_emd_path(name_or_path))
        return super().load(name_or_path, strict=False)

    def save(
        self,
        name_or_path,
        framework="PyTorch",
        publish=False,
        gis=None,
        compute_metrics=True,
        save_optimizer=False,
        **kwargs,
    ):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by :class:`~arcgis.learn.SingleShotDetector`, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
                                Boolean `zip_files` if True, it will create the Deep
                                Learning Package (DLPK) file while saving the model.
        =====================   ===========================================

        :return: the qualified path at which the model is saved
        """

        from ..models._arcgis_model import _create_zip

        zip_files = kwargs.pop("zip_files", True)
        overwrite = kwargs.pop("overwrite", False)
        if "\\" in name_or_path or "/" in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(self._data.path, "models", name_or_path)
            if not os.path.exists(os.path.dirname(path)):
                os.mkdir(os.path.dirname(path))

        path = super().save(
            path,
            framework,
            publish=False,
            gis=None,
            compute_metrics=compute_metrics,
            save_optimizer=save_optimizer,
            zip_files=False,
            **kwargs,
        )

        self._save_df_to_html(path)

        if zip_files:
            _create_zip(path.name, str(path))

        if publish:
            self._publish_dlpk(
                (path / path.stem).with_suffix(".dlpk"), gis=gis, overwrite=overwrite
            )

        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io

        with io.capture_output() as captured:
            metrics = self.get_model_metrics()

        return metrics

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        # metrics = self.get_model_metrics()
        # _emd_template.update(metrics)
        is_multilabel_problem = True if len(self._data._label_cols) > 1 else False
        _emd_template["Architecture"] = self.learn.model._transformer_architecture
        _emd_template[
            "PretrainedModel"
        ] = self.learn.model._transformer_pretrained_model_name
        _emd_template["ModelType"] = "Transformer"
        _emd_template["MixedPrecisionTraining"] = self._mixed_precision
        _emd_template["TextColumns"] = self._data._text_cols
        _emd_template["LabelColumns"] = self._data._label_cols
        _emd_template["SequenceLength"] = self._seq_len
        return _emd_template

    def show_results(self, rows=5, **kwargs):
        """
        Prints the rows of the dataframe with target and prediction columns.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================

        :return: dataframe
        """
        self._check_requisites()
        rows = (
            rows if (rows <= self.learn.data.batch_size) else self.learn.data.batch_size
        )
        return self.learn.show_results(rows=rows, **kwargs)

    def get_model_metrics(self):
        """
        Calculates the following  metrics:

        * accuracy:   the number of correctly predicted labels in the validation set divided by the total number of items in the validation set

        * bleu-score  This value indicates the similarity between model predictions and the ground truth text. Maximum value is 1

        :return: a dictionary containing the metrics for classification model.
        """
        try:
            self._check_requisites()
        except Exception as e:
            acc, bleu = self._data.emd.get("seq2seq_acc"), self._data.emd.get("bleu")
            if acc or bleu:
                return {"seq2seq_acc": acc, "bleu": bleu}
            else:
                self.logger.error("Metric not found in the loaded model")
        else:
            if hasattr(self.learn, "recorder"):
                metrics_names = self.learn.recorder.metrics_names
                metrics_values = self.learn.recorder.metrics
                if len(metrics_names) > 0 and len(metrics_values) > 0:
                    metrics = {
                        x: round(float(metrics_values[-1][i]), 4)
                        for i, x in enumerate(metrics_names)
                    }
                else:
                    metrics = self._calculate_model_metrics()
            else:
                metrics = self._calculate_model_metrics()
            return metrics

    def _calculate_model_metrics(self):
        self._check_requisites()
        self.logger.info("Calculating Model Metrics")
        metrics_names = ["accuracy", "bleu"]
        metrics_values = self.learn.validate()[1:]  # 0th value is validation loss
        metrics = {}
        if len(metrics_names) > 0 and len(metrics_values) > 0:
            metrics = {
                x: round(float(metrics_values[i]), 4)
                for i, x in enumerate(metrics_names)
            }
        return metrics

    def predict(
        self,
        text_or_list,
        batch_size=64,
        show_progress=True,
        explain=False,
        explain_index=None,
        **kwargs,
    ):
        """
        Predicts the translated outcome.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required input string or list of input strings.
        ---------------------   -------------------------------------------
        batch_size              Optional integer.
                                Number of inputs to be processed at once.
                                Try reducing the batch size in case of out of
                                memory errors.
                                Default value : 64
        ---------------------   -------------------------------------------
        show_progress           Optional bool.
                                To show or not to show the progress of prediction task.
                                Default value : True
        ---------------------   -------------------------------------------
        explain                 Optional bool.
                                To enable shap based importance
                                Default value : False
        ---------------------   -------------------------------------------
        explain_index           Optional list.
                                Index of the input rows for which the importance score will
                                be generated
                                Default value : None
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        num_beams               Optional integer.
                                Number of beams for beam search. 1 means no beam search.
                                Default value is set to 1
        ---------------------   -------------------------------------------
        max_length              Optional integer.
                                The maximum length of the sequence to be generated.
                                Default value is set to 20
        ---------------------   -------------------------------------------
        min_length              Optional integer.
                                The minimum length of the sequence to be generated.
                                Default value is set to 10
        =====================   ===========================================

        :return: list of tuples(input , predicted output strings).
        """

        if isinstance(text_or_list, str):
            text_or_list = [text_or_list]
        preds = self.learn.predict(text_or_list, batch_size, show_progress, **kwargs)

        # select the specific rows for the explanation
        text_list_for_exp = []
        try:
            if explain:
                if isinstance(explain_index, list):
                    # validate the index
                    temp_index = []
                    invalid_index = []
                    for index in explain_index:
                        if isinstance(index, int):
                            temp_index.append(index)
                        else:
                            invalid_index.append(index)

                    if invalid_index:
                        with warnings.catch_warnings():
                            warnings.simplefilter("always", UserWarning)
                            warnings.warn(
                                f"Index {invalid_index} are not valid. Indices/index must be integer. Ignoring "
                                f"{invalid_index} for processing."
                            )

                    if temp_index:
                        for i in temp_index:
                            if i < len(text_or_list):
                                text_list_for_exp.append(text_or_list[i])
                            else:
                                with warnings.catch_warnings():
                                    warnings.simplefilter("always", UserWarning)
                                    warnings.warn(
                                        f"Value of index {i} should be less than/equal to {len(text_or_list) -1}."
                                    )
                    else:
                        with warnings.catch_warnings():
                            warnings.simplefilter("always", UserWarning)
                            warnings.warn(
                                f"No valid indices were supplied. Please change your input to list of integers"
                            )

                elif isinstance(explain_index, int):
                    if explain_index < len(text_or_list):
                        text_list_for_exp = text_or_list[explain_index]
                    else:
                        with warnings.catch_warnings():
                            warnings.simplefilter("always", UserWarning)
                            warnings.warn(
                                f"Value of index {explain_index} should be less than/equal to {len(text_or_list) - 1}."
                            )
                else:
                    exp_rows = 5 if len(text_or_list) > 5 else len(text_or_list)
                    for i in range(exp_rows):
                        text_list_for_exp.append(text_or_list[i])
                    with warnings.catch_warnings():
                        warnings.simplefilter("always", UserWarning)
                        warnings.warn(
                            f"Generating explanation for first {exp_rows} rows "
                        )

            if explain and len(text_list_for_exp):
                self._explain(text_list_for_exp, **kwargs)

        except:
            with warnings.catch_warnings():
                warnings.simplefilter("always", UserWarning)
                warnings.warn(
                    f"SHAP workflow has encountered an error. Failed to generate an explanation."
                )

        return list(zip(text_or_list, preds))

    def _save_df_to_html(self, path):
        if getattr(self._data, "_is_empty", False):
            copy_metrics(self._data.emd_path, path, model_characteristics_folder)
            return
        validation_dataframe = self._data._valid_df.sample(n=5)

        model_output = self.predict(
            validation_dataframe[self._data._text_cols].tolist(), show_progress=False
        )
        predictions = [pred for _, pred in model_output]
        labels = [
            x[0] for x in validation_dataframe[self._data._label_cols].values.tolist()
        ]
        new_df = pd.DataFrame(
            validation_dataframe[self._data._text_cols].values, columns=["input"]
        )
        new_df["target"] = labels
        new_df["predictions"] = predictions

        df_str = new_df.to_html(index=False, justify="left").replace(">\n", ">")

        msg = "<p><b>Sample Results</b></p>"

        text = f"\n\n{msg}\n\n{df_str}"

        save_data_in_model_metrics_html(text, path, model_characteristics_folder)

    def plot_losses(self, show=True):
        """
        Plot training and validation losses.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        show                    Optional bool. Defaults to True
                                If set to False, figure will not be plotted
                                but will be returned, when set to True function
                                will plot the figure and return nothing.
        =====================   ===========================================

        :return: `matplotlib.figure.Figure <https://matplotlib.org/stable/api/figure_api.html#matplotlib.figure.Figure>`_
        """
        self._check_requisites()
        import matplotlib.pyplot as plt

        if not hasattr(self.learn, "recorder"):  # return none if the recorder is empty
            self.logger.error(
                "Model needs to be trained first. Please call `model.fit()` to train the model."
                " Then call this method to plot/return the loss curve."
            )
            return
        return_fig = not show
        fig = self.learn.recorder.plot_losses(return_fig=return_fig)
        if return_fig:
            plt.close()
            return fig

    def _explain(self, text_list_for_exp, **kwargs):
        """
        Generates SHAP based explanation.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_list_for_exp       Required input string or list of input strings.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        num_beams               Optional integer.
                                Number of beams for beam search. 1 means no beam search.
                                Default value is set to 1
        ---------------------   -------------------------------------------
        max_length              Optional integer.
                                The maximum length of the sequence to be generated.
                                Default value is set to 20
        ---------------------   -------------------------------------------
        min_length              Optional integer.
                                The minimum length of the sequence to be generated.
                                Default value is set to 10
        =====================   ===========================================

        :return: list of tuples(input , predicted output strings).
        """
        has_shap = True
        try:
            import shap
        except:
            has_shap = False
            warnings.warn(
                "SHAP is not installed. Model explainablity will not be available"
            )
        if has_shap:
            if isinstance(text_list_for_exp, str):
                text_list_for_exp = [text_list_for_exp]

            # collect the model and tokenizer
            emodel = self.learn.model._transformer
            emask = self.learn.model._tokenizer

            ###
            # There is an issue with the fast tokenizer. So selecting the slow tokenizer
            ###
            if isinstance(emask, transformers.PreTrainedTokenizerFast):
                if emask.name_or_path.find("T5"):
                    emask = transformers.T5TokenizerFast.from_pretrained(
                        "google/t5-v1_1-base", from_slow=True
                    )

            # initialize the explainer.
            explainer = shap.Explainer(emodel, emask)

            # validated the kwargs. So that we can match maximum length as passed to inference model
            try:
                for key, val in kwargs.items():
                    if hasattr(explainer.masker.model.inner_model.config, key):
                        explainer.masker.model.inner_model.config.__dict__[key] = val
            except:
                pass

            # generate shap values and generate plot.
            for record in text_list_for_exp:
                self.shap_values = explainer([record], fixed_context=0)
                shap.plots.text(self.shap_values)

    # def summarize(self, text_or_list, num_beams=4, max_len=50):
    #     if self._model_type not in['t5']:
    #         return logger.warning(f'Function not implemented for {self._model_type} models')
    #     if isinstance(text_or_list, str):
    #        text_or_list = ['summarize : ' + text_or_list]
    #     elif isinstance(text_or_list, list):
    #         text_or_list = ['summarize : '+ x for x in text_or_list]
    #     return self.learn.predict(text_or_list)

    # def qna(self, question_text_or_list, context_text_or_list, num_beams=4, max_len=50):
    #     if self._model_type not in['t5']:
    #         return logger.warning(f'Function not implemented for {self._model_type} models')
    #     if type(question_text_or_list) != type(context_text_or_list):
    #         return('Questions and context must either be both of string type or equal length lists of strings.')
    #     if isinstance(question_text_or_list, str) and isinstance(context_text_or_list, str):
    #         question_text_or_list = 'question: ' + question_text_or_list
    #         context_text_or_list = ' context: ' + context_text_or_list
    #         text_or_list = [question_text_or_list + context_text_or_list]
    #     elif isinstance(question_text_or_list, list) and isinstance(context_text_or_list, list):
    #         question_text_or_list = ['question: ' + x for x in  question_text_or_list]
    #         context_text_or_list = [' context: ' + x for x in  context_text_or_list]
    #         text_or_list = [x+y for x,y in zip(question_text_or_list,context_text_or_list)]
    #     return self.learn.predict(text_or_list)



from __future__ import annotations
from arcgis.gis import GIS
from typing import Any, TypeVar

__all__ = ["Container", "ContainerManager"]


K = TypeVar("K")
V = TypeVar("V")


class Container:
    """
    Represents a Single Notebook Container

    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The url for the container.
    ----------------  -------------------------------------------------------------------------------
    gis               Required GIS. The ArcGIS Online connection object.
    ================  ===============================================================================


    """

    _properties = None

    def __init__(self, url: str, gis: GIS):
        """initalizer"""
        self._url = url
        self._gis = gis

    @property
    def properties(self):
        if self._properties is None:
            url = f"{self._url}"
            params = {"f": "json"}
            self._properties = self._gis._con.get(url, params)
        return self._properties

    def terminate(self) -> bool:
        """stops the current container"""
        url = f"{self._url}/terminateContainer"
        params = {"f": "json"}
        return self._gis._con.post(url, params)

    @property
    def notebooks(self) -> list[dict[str, Any]]:
        """returns a list of notebooks running in the current container"""
        url = f"{self._url}/notebooks"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    def close(self, notebook_id: str) -> bool:
        """closes a notebook"""
        url = f"{self._url}/notebooks/{notebook_id}/closeNotebook"
        params = {"f": "json"}
        return self._gis._con.post(url, params)


class ContainerManager:
    """
    ================  ===============================================================================
    **Parameter**      **Description**
    ----------------  -------------------------------------------------------------------------------
    url               Required String. The base url for the ContainerManager endpoints.
    ----------------  -------------------------------------------------------------------------------
    gis               Required GIS. The ArcGIS Online connection object.
    ================  ===============================================================================

    """

    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    def list(self) -> list[dict[str, Any]]:
        """Returns a list of containers"""
        url = f"{self._url}"
        params = {"f": "json"}
        return self._gis._con.get(url, params)

    def get(self, id: str) -> Container:
        """Gets an instance of a container"""
        return Container(url=f"{self._url}/{id}", gis=self._gis)

    def start(self, runtime: str, instance_type: str | None = None) -> dict[K, V]:
        """starts a container"""
        url = f"{self._url}/startContainer"
        params = {
            "f": "json",
            "notebookRuntimeId": runtime,
        }
        if instance_type:
            params["instanceTypeName"] = instance_type
        res = self._gis._con.post(url, params)
        return res



from __future__ import annotations
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
from arcgis.geocoding import geocode
from arcgis.apps.hub.initiatives import InitiativeManager, Initiative
from arcgis.apps.hub.events import EventManager, Event
from arcgis.apps.hub.sites import SiteManager, Site
from arcgis.apps.hub.pages import PageManager, Page
from datetime import datetime
from collections import OrderedDict
import json
from functools import wraps


def _lazy_property(fn):
    """Decorator that makes a property lazy-evaluated."""
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = "_lazy_" + fn.__name__

    @property
    @wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


class Hub(object):
    """
    Entry point into the Hub module. Lets you access an individual hub and its components.


    ================    ===============================================================
    **Parameter**        **Description**
    ----------------    ---------------------------------------------------------------
    url                 Required string. If no URL is provided by user while connecting
                        to the GIS, then the URL will be ArcGIS Online.
    ----------------    ---------------------------------------------------------------
    GIS                 Required authenticated GIS object for the ArcGIS Online
                        organization associated with your Hub.
    ================    ===============================================================

    """

    def __init__(self, gis: GIS):
        self.gis = gis
        try:
            self._gis_id = self.gis.properties.id
        except AttributeError:
            self._gis_id = None

    @property
    def _hub_enabled(self):
        """
        Returns True if Hub Premium is enabled on this org
        """
        try:
            self.gis.properties.portalProperties["hub"]["enabled"]
            return True
        except:
            return False

    @property
    def enterprise_org_id(self) -> str:
        """
        Returns the AGOL org id of the Enterprise Organization associated with this Premium Hub.
        """

        if self._hub_enabled:
            try:
                _e_org_id = (
                    self.gis.properties.portalProperties.hub.settings.enterpriseOrg.orgId
                )
                return _e_org_id
            except AttributeError:
                try:
                    if (
                        self.gis.properties.subscriptionInfo.companionOrganizations.type
                        == "Enterprise"
                    ):
                        return "Enterprise org id is not available"
                except:
                    return self._gis_id
        else:
            raise Exception("Hub does not exist or is inaccessible.")

    @property
    def community_org_id(self) -> str:
        """
        Returns the AGOL org id of the Community Organization associated with this Premium Hub.
        """
        if self._hub_enabled:
            try:
                _c_org_id = (
                    self.gis.properties.portalProperties.hub.settings.communityOrg.orgId
                )
                return _c_org_id
            except AttributeError:
                try:
                    if (
                        self.gis.properties.subscriptionInfo.companionOrganizations.type
                        == "Community"
                    ):
                        return "Community org id is not available"
                except:
                    return self._gis_id
        else:
            raise Exception("Hub does not exist or is inaccessible.")

    @property
    def enterprise_org_url(self) -> str:
        """
        Returns the AGOL org url of the Enterprise Organization associated with this Premium Hub.
        """
        if self._hub_enabled:
            try:
                self.gis.properties.portalProperties.hub.settings.enterpriseOrg
                try:
                    _url = self.gis.properties.publicSubscriptionInfo.companionOrganizations[
                        0
                    ][
                        "organizationUrl"
                    ]
                except:
                    _url = self.gis.properties.subscriptionInfo.companionOrganizations[
                        0
                    ]["organizationUrl"]
                return "https://" + _url
            except AttributeError:
                return self.gis.url
        else:
            raise Exception("Hub does not exist or is inaccessible.")

    @property
    def community_org_url(self) -> str:
        """
        Returns the AGOL org id of the Community Organization associated with this Premium Hub.
        """
        if self._hub_enabled:
            try:
                self.gis.properties.portalProperties.hub.settings.communityOrg
                try:
                    _url = self.gis.properties.publicSubscriptionInfo.companionOrganizations[
                        0
                    ][
                        "organizationUrl"
                    ]
                except:
                    _url = self.gis.properties.subscriptionInfo.companionOrganizations[
                        0
                    ]["organizationUrl"]
                return "https://" + _url
            except AttributeError:
                return self.gis.url
        else:
            raise Exception("Hub does not exist or is inaccessible.")

    @_lazy_property
    def initiatives(self):
        """
        The resource manager for Hub initiatives. See :class:`~arcgis.apps.hub.InitiativeManager`.
        """
        if self._hub_enabled:
            return InitiativeManager(self)
        else:
            raise Exception(
                "Initiatives are only available with Hub Premium. Please upgrade to Hub Premium to use this feature."
            )

    @_lazy_property
    def events(self):
        """
        The resource manager for Hub events. See :class:`~arcgis.apps.hub.EventManager`.
        """
        if self._hub_enabled:
            return EventManager(self)
        else:
            raise Exception(
                "Events is only available with Hub Premium. Please upgrade to Hub Premium to use this feature."
            )

    @_lazy_property
    def sites(self):
        """
        The resource manager for Hub sites. See :class:`~hub.sites.SiteManager`.
        """
        return SiteManager(self)

    @_lazy_property
    def pages(self):
        """
        The resource manager for Hub pages. See :class:`~hub.sites.PageManager`.
        """
        return PageManager(self)



import uuid
import base64
import socket
import hashlib
from requests import PreparedRequest, Response, Session
from arcgis.auth._auth._base import BaseEsriAuth
import datetime as _dt
import secrets
import hashlib
import base64
import requests
from arcgis.auth.tools import parse_url
from urllib.parse import parse_qs


###########################################################################
class EsriPKCEAuth(BaseEsriAuth):
    """Implements OAuth 2.0 PKCE Workflow"""

    _url = None
    _token = None
    _verify = None
    _proxies = None
    _referer = None
    _session = None
    _hostname = None
    _password = None
    _username = None
    _client_id = None
    _expiration = None
    _expires_in = None
    _tokens = None

    # ---------------------------------------------------------------------
    def __init__(
        self,
        url: str,
        username: str,
        password: str,
        *,
        legacy: bool = False,
        **kwargs,
    ):
        """initializer"""
        self._no_go_token = set()
        self.legacy = legacy
        self._hostname = kwargs.pop("hostname", socket.getfqdn().lower())

        self._redirect_url = kwargs.pop("redirect_url", url)
        self._client_id = kwargs.pop("client_id", "arcgisonline")
        self._referer = kwargs.pop("referer", "http")
        self._expiration = kwargs.pop("expiration", 1440)
        self._username = username
        self._url = url
        self._password = password
        self._verify = kwargs.pop("verify_cert", True)
        self._proxies = kwargs.pop("proxies", None)
        self._code_verifier = (uuid.uuid4().hex + uuid.uuid4().hex)[:44]
        self._code_challenge = None
        self._session = Session()
        self._session.proxies = self._proxies
        self._session.verify = self._verify

    # ---------------------------------------------------------------------
    @property
    def _create_challenge(self) -> str:
        """creates the challenge string for the PKCE login"""
        if self._code_challenge is None:
            self._code_challenge = (
                base64.urlsafe_b64encode(
                    hashlib.sha256(self._code_verifier.encode("ascii")).digest()
                )
                .decode()
                .strip("=")
            )
        return self._code_challenge

    # ---------------------------------------------------------------------
    def _signin(self, username: str, password: str) -> dict:
        """Signs into the enterprise or online site"""

        params = {
            "client_id": self._client_id,
            "redirect_uri": self._redirect_url,
            "response_type": "code",
            "code_challenge": self._create_challenge,
            "code_challenge_method": "S256",
            "f": "json",
            "expiration": self._expiration,
        }
        url = f"{self._url}/sharing/rest/oauth2/authorize"
        response = self._session.get(
            url, params=params, verify=self._verify, proxies=self._proxies
        ).text
        if "oauth_state" in response:
            oauth_state = response.split('"oauth_state":"')[1].split('"')[0]
        else:
            raise Exception("Unable to generate oauth token")
        params = {
            "oauth_state": oauth_state,
            "username": username,
            "password": password,
        }
        url = f"{self._url}/sharing/rest/oauth2/signin"
        response = self._session.post(
            url,
            data=params,
            verify=self._verify,
            proxies=self._proxies,
            allow_redirects=False,
        )
        if response.headers["Location"].lower().find("accepttermsandconditions") > -1:
            url: str = response.headers["Location"]
            params = {"acceptTermsAndConditions": True}
            response: requests.Response = self._session.post(
                url,
                data=params,
                verify=self._verify,
                proxies=self._proxies,
                allow_redirects=False,
            )

        if response.status_code == 302:
            if "Location" in response.headers:
                oauth_code = (
                    response.headers["Location"].split("code=")[1].split("&")[0]
                )

            else:
                raise Exception("Unable to generate oauth authorization code")
        else:
            raise Exception("Unable to generate oauth token\n{}".format(response))

        params = {
            "client_id": self._client_id,
            "grant_type": "authorization_code",
            "redirect_uri": self._url,
            "code_verifier": self._code_verifier,
            "code": oauth_code,
        }
        url = f"{self._url}/sharing/rest/oauth2/token"

        tokens = self._session.post(
            url, data=params, verify=self._verify, proxies=self._proxies
        ).json()
        self._expires_in = _dt.datetime.now() + _dt.timedelta(
            seconds=tokens.get("expires_in", 1800) - 300
        )
        self._refresh_token_value = tokens.get("refresh_token")
        self._refresh_expires_in = _dt.datetime.now() + _dt.timedelta(
            seconds=(tokens.get("refresh_token_expires_in", 86300) / 60) - 300
        )
        if "error" in tokens.keys():
            if "message" in tokens["error"]:
                raise Exception(
                    "Error generating access token\n{}".format(
                        tokens["error"]["message"]
                    )
                )
            else:
                raise Exception("Error generating access token")

        self._tokens = tokens

    # ---------------------------------------------------------------------
    def _refresh_token(self) -> None:
        """
        refreshes the `token` by using the refresh token
        This method updates the self._tokens value with the new access_token
        :returns: None
        """
        if _dt.datetime.now() >= self._refresh_expires_in:
            self._signin(username=self._username, password=self._password)
        params = {
            "client_id": self._client_id,
            "grant_type": "refresh_token",
            "refresh_token": self._refresh_token_value,
        }
        url = f"{self._url}/sharing/rest/oauth2/token"
        response = self._session.post(
            url, data=params, verify=self._verify, proxies=self._proxies
        )
        tokens = response.json()
        self._expires_in = _dt.datetime.now() + _dt.timedelta(
            seconds=tokens.get("expires_in", 1800) - 300
        )
        self._tokens.update(tokens)

    # ---------------------------------------------------------------------
    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        """
        Called Before PrepareRequest is Completed.

        The logic to attached preemptive authentication should go here.
        If you want to wait for response, register a response_hook within the call.

        :returns: PrepareRequest

        ```
        def __call__(self, r:PreparedRequest) -> PreparedRequest:
            r.register_hook(event="response", hook=self.response_hook)
            return r
        ```

        """
        if self.legacy == False:
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
        elif self.legacy and r.method == "GET":
            r.prepare_url(url=r.url, params={"token": self.token})
        elif self.legacy and r.method == "POST":
            data = parse_qs(r.body)
            data["token"] = self.token
            r.prepare_body(data, None, None)
        else:
            raise Exception(
                "Only POST and GET are supported with legacy methods of authentication."
            )
        r.register_hook("response", self.response_hook)
        try:
            self.pos = r.body.tell()
        except AttributeError:
            # In the case of HTTPKerberosAuth being reused and the body
            # of the previous request was a file-like object, pos has
            # the file position of the previous body. Ensure it's set to
            # None.
            self.pos = None
        # return request

        return r

    # ---------------------------------------------------------------------
    def response_hook(self, r: Response, **kwargs) -> Response:
        """
        response hook logic

        :return: Response
        """
        parsed = parse_url(r.url)
        if (
            r.text.lower().find("invalid token") > -1
            or r.text.find("Token is valid but access is denied.") > -1
            or (parsed.scheme, parsed.netloc, parsed.path) in self._no_go_token
        ):
            # parsed = parse.urlparse(r.url)
            self._no_go_token.add((parsed.scheme, parsed.netloc, parsed.path))
            # Recreate the request without the token
            #
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self._referer  # or "http"
            r.request.headers.pop("X-Esri-Authorization", None)
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer  # or "http"
            _r.headers.pop("X-Esri-Authorization", None)
            _r.history.append(r)
            return _r
        elif r.text.lower().find("token required") > -1:
            r.content
            r.raw.release_conn()
            r.request.headers["referer"] = self._referer  # or "http"
            r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
            _r = r.connection.send(r.request, **kwargs)
            _r.headers["referer"] = self._referer  # or "http"
            _r.headers["X-Esri-Authorization"] = f"Bearer {self.token}"
            _r.history.append(r)
            return _r
        return r

    # ---------------------------------------------------------------------
    @property
    def token(self) -> str:
        """
        returns a ArcGIS Token as a string

        :return: string
        """
        if self._tokens is None:
            self._signin(username=self._username, password=self._password)
            return self.token
        elif _dt.datetime.now() >= self._expires_in:
            self._refresh_token()
            return self.token
        elif _dt.datetime.now() >= self._refresh_expires_in:
            self._signin(username=self._username, password=self._password)
            return self.token
        else:
            return self._tokens.get("access_token")



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# error in show_result due thier bug
# grid_rcnn_r50_fpn_gn-head_groie_1x_coco.py, box AP=39.1

_base_ = "./_base_/grid_rcnn_r50_fpn_gn-head_2x_coco.py"
model = dict(
    roi_head=dict(
        bbox_roi_extractor=dict(
            type="GenericRoIExtractor",
            aggregation="sum",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=2),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
            pre_cfg=dict(
                type="ConvModule",
                in_channels=256,
                out_channels=256,
                kernel_size=5,
                padding=2,
                inplace=False,
            ),
            post_cfg=dict(
                type="GeneralizedAttention",
                in_channels=256,
                spatial_range=-1,
                num_heads=6,
                attention_type="0100",
                kv_stride=2,
            ),
        ),
        grid_roi_extractor=dict(
            type="GenericRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=14, sampling_ratio=2),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
            pre_cfg=dict(
                type="ConvModule",
                in_channels=256,
                out_channels=256,
                kernel_size=5,
                padding=2,
                inplace=False,
            ),
            post_cfg=dict(
                type="GeneralizedAttention",
                in_channels=256,
                spatial_range=-1,
                num_heads=6,
                attention_type="0100",
                kv_stride=2,
            ),
        ),
    )
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/groie/grid_rcnn_r50_fpn_gn-head_1x_coco/grid_rcnn_r50_fpn_gn-head_1x_coco_20200605_202059-64f00ee8.pth"



# MIT License

# Copyright (c) 2018 XuanyiLi

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Based on https://github.com/meteorshowers/hed-pytorch/

import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import torch
from fastai.callbacks.hooks import hook_outputs
from fastai.callbacks.hooks import model_sizes
from fastai.vision.models.unet import _get_sfs_idxs
from fastprogress.fastprogress import progress_bar
from fastai.vision import flatten_model
from ._timm_utils import get_backbone
from fastai.basic_train import LearnerCallback
from torch.nn.parallel import DistributedDataParallel


def modify_layers(backbone, backbone_fn):
    hookable_modules = flatten_model(backbone)
    for i, module in enumerate(hookable_modules):
        if (
            isinstance(module, nn.MaxPool2d)
            and not "xception" in backbone_fn.__module__
        ):
            module.ceil_mode = True
            module.kernel_size = 2
        elif isinstance(module, nn.Conv2d) and i == 0:
            module.stride = (1, 1)


def get_hooks(backbone, chip_size):
    try:
        hookable_modules = flatten_model(backbone)
        feature_sizes = model_sizes(
            nn.Sequential(*hookable_modules), size=(chip_size, chip_size)
        )
        f_change_idxs = _get_sfs_idxs(feature_sizes)
        hooks = [hookable_modules[i] for i in f_change_idxs + [-1]][:5]
        return hooks
    except:
        hookable_modules = list(backbone.children())
        feature_sizes = model_sizes(
            nn.Sequential(*hookable_modules), size=(chip_size, chip_size)
        )
        f_change_idxs = _get_sfs_idxs(feature_sizes)
        if len(f_change_idxs) > 1 and f_change_idxs[0] == f_change_idxs[1]:
            del f_change_idxs[0]
        if len(f_change_idxs) > 3:
            hooks = [hookable_modules[i] for i in f_change_idxs + [-1]][:5]
            return hooks

        def get_len(m):
            try:
                return len(m)
            except:
                return 0

        module_len = [get_len(m) for m in hookable_modules]
        hooks = []
        idx = 0
        while len(hooks) < 5 and len(hookable_modules) > idx:
            if module_len[idx] == 0:
                try:
                    if module_len[idx + 1] > 0:
                        hooks.append(hookable_modules[idx])
                except:
                    hooks.append(hookable_modules[idx])
            else:
                feature_sizes = model_sizes(
                    nn.Sequential(*hookable_modules[idx]), size=(chip_size, chip_size)
                )
                if len(feature_sizes) < 2:
                    hooks.append(hookable_modules[idx])
                else:
                    f_change_idxs = _get_sfs_idxs(feature_sizes)
                    if len(f_change_idxs) > 1 and f_change_idxs[0] == f_change_idxs[1]:
                        del f_change_idxs[0]
                    if f_change_idxs == []:
                        hooks.append(hookable_modules[idx])
                    int_idx = 0
                    while len(hooks) < 5 and len(f_change_idxs) > int_idx:
                        hooks.append(hookable_modules[idx][f_change_idxs[int_idx]])
                        int_idx += 1
            idx += 1
        if len(hooks) < 5:
            hooks.append(hookable_modules[idx - 1][-1])
        return hooks


class _HEDModel(nn.Module):
    def __init__(self, backbone_fn, chip_size=224, pretrained=True):
        super().__init__()
        self.backbone = get_backbone(backbone_fn, pretrained)
        modify_layers(self.backbone, backbone_fn)
        if len(self.backbone) < 2:
            self.backbone = self.backbone[0]
        hooks = get_hooks(self.backbone, chip_size)
        self.hook = hook_outputs(hooks)
        model_sizes(self.backbone, size=(chip_size, chip_size))
        layer_num_channels = [k.stored.shape[1] for k in self.hook]

        self.score_dsn1 = nn.Conv2d(layer_num_channels[0], 1, 1)
        self.score_dsn2 = nn.Conv2d(layer_num_channels[1], 1, 1)
        self.score_dsn3 = nn.Conv2d(layer_num_channels[2], 1, 1)
        self.score_dsn4 = nn.Conv2d(layer_num_channels[3], 1, 1)
        self.score_dsn5 = nn.Conv2d(layer_num_channels[4], 1, 1)
        self.score_final = nn.Conv2d(5, 1, 1)

    def forward(self, x):
        img_H, img_W = x.shape[2], x.shape[3]
        x = self.backbone(x)
        features = self.hook.stored

        so1 = self.score_dsn1(features[0])
        so2 = self.score_dsn2(features[1])
        so3 = self.score_dsn3(features[2])
        so4 = self.score_dsn4(features[3])
        so5 = self.score_dsn5(features[4])

        weight_deconv2 = make_bilinear_weights(4, 1).to(x.device)
        weight_deconv3 = make_bilinear_weights(8, 1).to(x.device)
        weight_deconv4 = make_bilinear_weights(16, 1).to(x.device)
        weight_deconv5 = make_bilinear_weights(32, 1).to(x.device)

        upsample2 = torch.nn.functional.conv_transpose2d(so2, weight_deconv2, stride=2)
        upsample3 = torch.nn.functional.conv_transpose2d(so3, weight_deconv3, stride=4)
        upsample4 = torch.nn.functional.conv_transpose2d(so4, weight_deconv4, stride=8)
        upsample5 = torch.nn.functional.conv_transpose2d(so5, weight_deconv5, stride=16)

        so2 = crop(upsample2, img_H, img_W)
        so3 = crop(upsample3, img_H, img_W)
        so4 = crop(upsample4, img_H, img_W)
        so5 = crop(upsample5, img_H, img_W)

        if so1.size(2) != so2.size(2):
            so1 = F.interpolate(
                so1, (img_H, img_W), mode="bilinear", align_corners=False
            )

        fusecat = torch.cat((so1, so2, so3, so4, so5), dim=1)
        fuse = self.score_final(fusecat)

        results = [so1, so2, so3, so4, so5, fuse]
        results = [torch.sigmoid(r) for r in results]

        return results


def crop(variable, th, tw):
    h, w = variable.shape[2], variable.shape[3]
    x1 = int(round((w - tw) / 2.0))
    y1 = int(round((h - th) / 2.0))
    return variable[:, :, y1 : y1 + th, x1 : x1 + tw]


def make_bilinear_weights(size, num_channels):
    factor = (size + 1) // 2
    if size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = np.ogrid[:size, :size]
    filt = (1 - abs(og[0] - center) / factor) * (1 - abs(og[1] - center) / factor)
    filt = torch.from_numpy(filt)
    w = torch.zeros(num_channels, num_channels, size, size)
    w.requires_grad = False
    for i in range(num_channels):
        for j in range(num_channels):
            if i == j:
                w[i, j] = filt
    return w


def cross_entropy_loss(prediction, label):
    label = label.long()
    mask = (label != 0).float()
    num_positive = torch.sum(mask).float()
    num_negative = mask.numel() - num_positive
    mask[mask != 0] = num_negative / (num_positive + num_negative)
    mask[mask == 0] = num_positive / (num_positive + num_negative)
    cost = torch.nn.functional.binary_cross_entropy(
        prediction.float(), label.float(), weight=mask
    )  # , reduce=False)
    return torch.sum(cost)


def hed_loss(out, labels):
    loss = 0.0
    for o in out:
        loss = loss + cross_entropy_loss(o, labels)
    return loss


def accuracy(input, target):
    if isinstance(input, tuple):  # while training
        input = input[0]
    target = target.byte().squeeze(1)
    input = input[-1]
    input = (input >= 0.5).byte().squeeze(1)
    return (input == target).float().mean()


def get_true_positive(mask1, mask2, buffer):
    tp = 0
    indices = np.where(mask1 == 1)
    for ind in range(len(indices[0])):
        tp += np.any(
            mask2[
                max(indices[0][ind] - buffer, 0) : indices[0][ind] + buffer + 1,
                max(indices[1][ind] - buffer, 0) : indices[1][ind] + buffer + 1,
            ]
        ).astype(np.int)
    return tp


def get_confusion_metric(gt, pred, buffer):
    from skimage.morphology import skeletonize, binary_dilation

    tp, predicted_tp, actual_tp = 0, 0, 0
    for i in range(gt.shape[0]):
        gt_mask = skeletonize(binary_dilation(gt[i]))
        pred_mask = skeletonize(binary_dilation(pred[i]))
        tp += get_true_positive(gt_mask, pred_mask, buffer)
        predicted_tp += len(np.where(pred_mask == 1)[0])
        actual_tp += len(np.where(gt_mask == 1)[0])

    return tp, predicted_tp, actual_tp


def f1_score(pred, gt):
    device = gt.device
    gt = gt.byte().squeeze(1).cpu().numpy()
    pred = (pred[-1] >= 0.5).byte().squeeze(1).cpu().numpy()
    tp, predicted_tp, actual_tp = get_confusion_metric(gt, pred, 3)
    precision = tp / (predicted_tp + 1e-12)
    recall = tp / (actual_tp + 1e-12)
    f1score = 2 * precision * recall / (precision + recall + 1e-12)
    return torch.tensor(f1score).to(device)


def accuracies(model, dl, detect_thresh=0.5, buffer=3, show_progress=True):
    precision, recall, f1score = [], [], []
    model.learn.model.eval()
    acc = {}
    with torch.no_grad():
        for input, gt in progress_bar(dl, display=show_progress):
            predictions = model.learn.model(input)
            gt = gt.byte().squeeze(1).cpu().numpy()
            pred = (predictions[-1] >= detect_thresh).byte().squeeze(1).cpu().numpy()
            tp, predicted_tp, actual_tp = get_confusion_metric(gt, pred, buffer)
            prec = tp / (predicted_tp + 1e-12)
            rec = tp / (actual_tp + 1e-12)
            precision.append(prec)
            recall.append(rec)
            f1score.append(2 * prec * rec / (prec + rec + 1e-12))
    acc["Precision"] = np.mean(precision)
    acc["Recall"] = np.mean(recall)
    acc["F1 Score"] = np.mean(f1score)

    return acc


class DDPCallback(LearnerCallback):
    def __init__(self, learn, cuda_id):
        super().__init__(learn)
        self.cuda_id = cuda_id

    def on_train_begin(self, **kwargs):
        self.learn.model = DistributedDataParallel(
            self.learn.model.module,
            device_ids=[self.cuda_id],
            output_device=self.cuda_id,
            find_unused_parameters=True,
        )
